<!DOCTYPE html>
<html>
	<head>
    <meta charset="utf-8" />
    <link rel="shortcut icon" href="/favicon.ico" type="image/x-icon">
		<title>Meerschaum API</title>
    <meta name="viewport" content="width=device-width, initial-scale=1">
    <link rel="stylesheet" href="https://maxcdn.bootstrapcdn.com/bootstrap/3.3.7/css/bootstrap.min.css" integrity="sha384-BVYiiSIFeK1dGmJRAkycuHAHRg32OmUcww7on3RYdg4Va+PmSTsz/K68vbdEjh4u" crossorigin="anonymous">
    <script src="https://ajax.googleapis.com/ajax/libs/jquery/3.5.1/jquery.min.js"></script>
    <link href="{{ url_for('static', path='/css/styles.css') }}" rel="stylesheet">
    <script src="{{ url_for('static', path='/js/main.js') }}"></script>
    <script></script>
	</head>
	<body>
    <div id="action_button_container"></div>
    <!-- <div class="container-fluid"> -->
      <!-- <button class="btn btn-default" onclick="click()">Click Me</button> -->
    <!-- </div> -->
    <div class="container-fluid">
      <!-- <label for="action">Choose an action:</label> -->
      <select class="form-select" aria-label="Default select example" name="action" id="action">
        <option selected>memes</option>
        <optgroup label="test">
          <option>test</option>
        </optgroup>
      </select>
    </div>

    <!-- Load React. -->
    <!-- Note: when deploying, replace "development.js" with "production.min.js". -->
    <script src="https://unpkg.com/react@17/umd/react.development.js" crossorigin></script>
    <script src="https://unpkg.com/react-dom@17/umd/react-dom.development.js" crossorigin></script>
    <!-- Load our React component. -->
    <script src="{{ url_for('static', path='/js/action_button.js') }}"></script>
	</body>
</html>
