a {
  color: var(--bs-yellow);
}
.btn-link {
  color: var(--bs-light);
}
.content-div-right pre {
  font-family: Consolas, Menlo, Monaco, Lucida Console, Liberation Mono, DejaVu Sans Mono, Bitstream Vera Sans Mono, Courier New, monospace, serif;
  margin-bottom: 10px;
  overflow: auto;
  width: 100ch;
  padding: 5px;
  width: 650px!ie7;
  padding-bottom: 20px!ie7;
  max-height: 600px;
}
#console-div {
  overflow-y: scroll;
  max-height: 85vh;
  height: 84vh;
  color: white;
  background-color: black;
  padding: 0.5em;
}
#webterm-iframe {
  width: 102%;
  height: 85vh !important;
  max-height: 85vh;
}
.codeblock {
  background-color: var(--bs-secondary) !important;
  color: white !important;
  display: block !important;
  overflow-x: auto !important;
  padding: 0.5em !important;

}
.running-job {
  color: limegreen;
}
.paused-job {
  color: yellow;
}
.stopped-job {
  color: tomato;
}
.plugin-description {
  border: 0px solid transparent !important;
  outline: none !important;
  margin: 0px;
  padding: 0px;
  display: block;

}
.plugin-description:focus, .plugin-description:active {
  outline: 0px solid transparent !important;
  border-color: transparent;
}
.job-card-text {
  font-family: Consolas, Menlo, Monaco, Lucida Console, Liberation Mono, DejaVu Sans Mono, Bitstream Vera Sans Mono, Courier New, monospace, serif;
}
.accordion {
  /* background-color: var(--dark); */
}
#get-items-menu {
  width: 100%;
}
#instance-select {
  background-color : var(--bs-secondary) !important;
  color: white;
  margin-top: 5px;
}
.input-text {
  font-family: Consolas, Menlo, Monaco, Lucida Console, Liberation Mono, DejaVu Sans Mono, Bitstream Vera Sans Mono, Courier New, monospace, serif !important;
}
